# -*- coding: utf-8 -*-
"""Stone Soup framework: development and assessment of tracking algorithms."""
__version__ = '0.1b1'
__copyright__ = '''\
Crown Copyright (c) 2017-2019 Defence Science and Technology Laboratory UK
Crown Copyright (c) 2018-2019 Defence Research and Development Canada / Recherche et développement pour la défense Canada
Copyright (c) 2018-2019 University of Liverpool UK
'''  # noqa:E501
__license__ = 'MIT'
